/*
 * Copyright (C) 1995, 1996  Warren Jasper
 * All rights reserved.
 *
 */

/***************************************************************************
 *
 * d2ac.c
 *
 ***************************************************************************/

#define MODULE

/* retrieve the CONFIG_* macros */
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include "dda06-16.h"
#include "d2ac.h"

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

static ssize_t dac_read(struct file *filePtr, char *buf, size_t count, loff_t *off);
static ssize_t dac_write(struct file *filePtr, const char *buf, size_t count, loff_t *off);
static int dac_open(struct inode *iNode, struct file *filePtr);
static int dac_close(struct inode *iNode, struct file *filePtr);
static int dac_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static int dac_find( WORD base_reg );
int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static WORD base_reg =  PORT_MIN;              /* base register address         */
static BYTE dio_reg = 0x80;                    /* status register 1             */
static int MajorNumber = DEFAULT_MAJOR_DEV;    /* Major number compiled in      */
static ChanRec Chan[DA_CHANNELS+DIO_PORTS];    /* Channel specific information  */

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static ssize_t dac_read(struct file *filePtr, char *buf, size_t count, loff_t *off)
{
  BYTE  value;
  int   minor;

  struct inode *iNode = filePtr->f_dentry->d_inode;
  minor = MINOR(iNode->i_rdev); 

  if ( minor >= 0 && minor < DA_CHANNELS ) {
      put_user(Chan[minor].value, (WORD*) buf);
      value = inb(Chan[minor].addr);
      return 2;
  }

  /* check to see if reading a value from the DIO */
  if ( minor >= DA_CHANNELS && minor < DA_CHANNELS + DIO_PORTS ) {
    value = inb(Chan[minor].addr);
    put_user(value, (BYTE*) buf);
    return 1;
  }

  printk("dac_read: Incorrect minor number (%d).\n", minor);
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static ssize_t dac_write(struct file *filePtr, const char *buf, size_t count, loff_t *off)
{
  int  minor;
  BYTE value;

  struct inode *iNode = filePtr->f_dentry->d_inode;
  minor = MINOR(iNode->i_rdev); 
  if ( minor >= 0 && minor < DA_CHANNELS ) {   /* write out D/A value */
      get_user(Chan[minor].value, (WORD*)buf);
      get_user(value, (BYTE*) buf);
      outb( value, Chan[minor].addr);
      get_user(value, (BYTE*) (buf+1));
      outb( value, Chan[minor].addr+1);
      return 1;                                /* return number of WORDS written */
  }

  /* check to see if reading a value from the DIO */
  if ( minor >= DA_CHANNELS && minor < DA_CHANNELS + DIO_PORTS ) {
    get_user(value, (BYTE *) buf);
    outb_p( value, Chan[minor].addr);
    
#ifdef DEBUG
    printk("%s: DIO Port %#x set to %#x\n", ADAPTER_ID, minor-DA_CHANNELS, value);
#endif
    return 1;
  }

  return -1;
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int dac_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);

  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
  */

  if ( Chan[minor].open == TRUE ) {
      return -EBUSY;
  }

  MOD_INC_USE_COUNT;

  Chan[minor].open = TRUE;              /* The device is open */

  #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, Chan[minor].mode);
  #endif
  return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static int dac_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);

  #ifdef DEBUG
      printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
  #endif

  MOD_DEC_USE_COUNT;
  Chan[minor].open = FALSE;
  return 0;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int dac_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  int err = 0;
  int size = _IOC_SIZE(cmd);       /* the size bitfield in cmd */

  /* 
   * extract the type and number bitfields, and don't decode
   * wrong cmds;  return EINVAL before verify_area()
   */

  if (_IOC_TYPE(cmd) != IOCTL_MAGIC) return -EINVAL;
  if (_IOC_NR(cmd) > IOCTL_MAXNR)  return -EINVAL;

  if (_IOC_DIR(cmd) & _IOC_READ) {
    err = verify_area(VERIFY_WRITE, (void *)arg, size);
  } else if (_IOC_DIR(cmd) & _IOC_WRITE) {
     err = verify_area(VERIFY_READ, (void *)arg, size);
  }
  if (err) return err;
  
  switch (cmd) {

    case DAC_XFER:
      /* update all channels at once */
      inb(base_reg);
      break;

    case DIO_SET_MODE:
      if (minor < DA_CHANNELS || minor > (DA_CHANNELS + DIO_PORTS))
        return(-EINVAL);
      arg &= 0x3;

      switch (minor - DA_CHANNELS) {
        case 0:			/* Port A */
          #ifdef DEBUG
          printk("DIO_SET_MODE for Port A\n");
          #endif
          dio_reg &= 0x9f;
          dio_reg |= (arg << 5);
          outb_p(dio_reg, DIO_CNTRL_REG);
          Chan[minor].mode = arg;
          break;

        case 1:			/* Port B */
          #ifdef DEBUG
          printk("DIO_SET_MODE for Port B\n");
          #endif
          if (arg & 0x2)
	    return(-EINVAL);	/* Port B only has Modes 0 & 1 */
          dio_reg &=  0xfb;
          dio_reg |= (arg << 2);
          outb_p(dio_reg, DIO_CNTRL_REG);
          Chan[minor].mode = arg;
          break;

        case 2:			/* Port C */
          #ifdef DEBUG
          printk("DIO_SET_MODE for Port C\n");
          #endif
          if (arg)
	    return(-EINVAL);	/* Port C only has Mode 0 I/O */
          Chan[minor].mode = arg;
          break;

        default:
          #ifdef DEBUG
          printk("DIO_SET_MODE for Invalid Port\n");
          #endif
          return(-EINVAL);	/* Wrong Port Number */
          break;
      }
    break;
    
    case DIO_SET_DIRECTION:
      #ifdef DEBUG
      printk("DIO_SET_DIRECTION: minor = %d, arg = %d\n", (int) minor, (int) arg );
      #endif

      if (minor < DA_CHANNELS || minor > (DA_CHANNELS + DIO_PORTS))
        return(-EINVAL);
    
      switch (minor - DA_CHANNELS) {
        case 0:			/* Port A */
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Port A\n");
          #endif
          arg &= 0x1;
          dio_reg &=  0xef;
          dio_reg |= (arg << 4);
          outb_p(dio_reg, DIO_CNTRL_REG);
          break;

        case 1:			/* Port B */
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Port B\n");
          #endif
          arg &= 0x1;
          dio_reg &=  0xfd;
          dio_reg |= (arg << 1);
          outb_p(dio_reg, DIO_CNTRL_REG);
          break;

        case 2:			/* Port C */
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Port C\n");
          #endif
          switch ( arg ) {
            case PORT_INPUT:         /* CU = IN  CL = IN */
                dio_reg |= 0x9;
                break;
            case PORT_OUTPUT:        /* CU = OUT  CL = OUT */
                dio_reg &= ~(0x9);
                break;
            case LOW_PORT_INPUT:     /* CL = IN */
                dio_reg |= 0x1;
                break;
            case LOW_PORT_OUTPUT:    /* CL = OUT */
                dio_reg &= ~(0x1);
                break;
            case HIGH_PORT_INPUT:    /* CU = IN */
                dio_reg |= 0x8;
                break;
            case HIGH_PORT_OUTPUT:   /* CU = OUT */
                dio_reg &= ~(0x8);
                break;
            default:
                return(-EINVAL);
                break;
          }
          outb_p(dio_reg, DIO_CNTRL_REG);
          break;

        default:
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Invalid Port\n");
          #endif
          return(-EINVAL);
          break;
        }
    break;
    
    default:
      return(-EINVAL);
      break;
  }
  return 0;
}

/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations dac_fops = {
	NULL,           /* dac_seek */
	dac_read,
	dac_write,      /* dac_write */
	NULL,		/* dac_readdir */
	NULL,		/* dac_select */
	dac_ioctl,	/* dac_ioctl */
	NULL,
	dac_open,
	NULL,
	dac_close,
	NULL		/* fsync */
};

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************
 *
 * Loads driver. Called when "insmod dda06-16.o" is invoked on the command line.
 *
 ***************************************************************************/

int init_module(void) 
{
  int i, err;

  /* Register as a device with kernel.  */

  if ((err = register_chrdev(MajorNumber, "dda06/16", &dac_fops))) {
      printk("%s: Failure to load module. error %d\n", ADAPTER_ID, -err);
      return err;
  } 

  /* Register the region of iospace with kernel and get base address */

#ifdef DAC_BASE
    base_reg = DAC_BASE;
    if ( check_region(base_reg, PORT_SIZE ) != 0 ) {  /* in use */
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, base_reg );
        return -ENODEV;
    } else {
      request_region( base_reg, PORT_SIZE, "dda06/16" );
    }
#else
  /* Look to see if DAC is installed */
  for ( base_reg = PORT_MIN; base_reg <= PORT_MAX; base_reg += PORT_STEP ) {
      if ( check_region(base_reg, PORT_SIZE ) == 0 ) {  /* not in use */
          request_region(base_reg, PORT_SIZE, "dda06/16");
          if (dac_find( base_reg )) {  /* found the board */
              break;
          } else {
              release_region(base_reg, PORT_SIZE);
          }
      }
  }

  if ( base_reg > PORT_MAX ) {
      printk("%s: No board found from address %#x to %#x.\n", 
             ADAPTER_ID, PORT_MIN, PORT_MAX);
          if (unregister_chrdev(MajorNumber, "dda06/16") != 0) {
              printk("%s: cleanup_module failed\n", ADAPTER_ID);
          }
      return -ENODEV;
  }

#endif

  printk("%s: address=%#x.", ADAPTER_ID, base_reg);
  printk(" 3/27/96 wjasper@tx.ncsu.edu\n");

  /* Set all channel structures to show nothing active/open */

  for (i = 0; i < DA_CHANNELS; i++) {
      Chan[i].open = FALSE;
      Chan[i].addr = base_reg + 2*i;
      Chan[i].mode = 0;
  }

  Chan[DA_CHANNELS].addr   = DIO_PORTA;
  Chan[DA_CHANNELS+1].addr = DIO_PORTB;
  Chan[DA_CHANNELS+2].addr = DIO_PORTC;

  for ( i = DA_CHANNELS; i < (DA_CHANNELS + DIO_PORTS); i++ ) {
      Chan[i].open = FALSE;
      Chan[i].mode = 0;
      outb_p(0x0, Chan[i].addr);
  }
  return 0;
}

/***************************************************************************
 *
 * Test memory locations for exsistance of dac board.
 *
 ***************************************************************************/
int dac_find( WORD base_reg)
{

/*
    To test if there is a DDA 06/16 board do the following:
        1. The first 12 registers are write only, so nothing should come
           back.
        2. The next three registers are the ports of the 8255.  Should
           be able to read and write from them.  The 8255 is in mode 0 on
           powerup
*/
    BYTE bReg, value;
    WORD i;
    value = 0xa5;

    for ( i = 0; i < 2*DA_CHANNELS; i++ ) {
        outb_p(value, base_reg + i);
        bReg = inb_p( base_reg + i);
        if ( bReg == value ) {
            return FALSE;   /* these are write only */
        }
    }

    /* Now write to the DIO ports */

    dio_reg = 0x80;
    outb_p(dio_reg, DIO_CNTRL_REG);   /* put in output mode 0 ? */

    for ( i = 2*DA_CHANNELS; i < (2*DA_CHANNELS + DIO_PORTS); i++ ) {
        outb_p(value, base_reg + i);
        bReg = inb_p(base_reg + i);
        if ( bReg != value ) {
            return FALSE;   /* not the DDA06/16 */
        }
    }
    return TRUE;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod dda06-16" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    return;
  }

  release_region( base_reg, PORT_SIZE );

  if (unregister_chrdev(MajorNumber, "dda06/16") != 0) {
    printk("%s: cleanup_module failed\n", ADAPTER_ID);
  } else {

  #ifdef DEBUG
      printk("%s: module removed.\n", ADAPTER_ID);
  #endif

  }    
}
